@@ -24,14 +24,20 @@ We represent this tree as a nested dictionary with the following form:
                    'children': {...}}}
     }
 """
+from __future__ import print_function, division, absolute_import
 
-
-from collections import defaultdict
+import bisect
+from collections import defaultdict, deque
 import linecache
-import itertools
+import sys
+import threading
+from time import sleep
+
 import toolz
 
-from .utils import format_time
+from .metrics import time
+from .utils import format_time, color_of, parse_timedelta
+from .compatibility import get_thread_identity
 
 
 def identifier(frame):
@@ -66,7 +72,7 @@ def info_frame(frame):
             'line': line}
 
 
-def process(frame, child, state, stop=None):
+def process(frame, child, state, stop=None, omit=None):
     """ Add counts from a frame stack onto existing state
 
     This recursively adds counts to the existing state dictionary and creates
@@ -86,9 +92,14 @@ def process(frame, child, state, stop=None):
      'description': 'root',
      'children': {'...'}}
     """
+    if omit is not None and any(frame.f_code.co_filename.endswith(o) for o in omit):
+        return False
+
     prev = frame.f_back
     if prev is not None and (stop is None or not prev.f_code.co_filename.endswith(stop)):
         state = process(prev, frame, state, stop=stop)
+        if state is False:
+            return False
 
     ident = identifier(frame)
 
@@ -190,7 +201,7 @@ def plot_data(state, profile_interval=0.010):
         try:
             colors.append(color_of(desc['filename']))
         except IndexError:
-            colors.append(palette[-1])
+            colors.append('gray')
 
         delta = (stop - start) / state['count']
 
@@ -202,6 +213,7 @@ def plot_data(state, profile_interval=0.010):
             x += width
 
     traverse(state, 0, 1, 0)
+    percentages = ["{:.2f}%".format(100 * w) for w in widths]
     return {'left': starts,
             'right': stops,
             'bottom': heights,
@@ -213,19 +225,133 @@ def plot_data(state, profile_interval=0.010):
             'line': lines,
             'line_number': line_numbers,
             'name': names,
-            'time': times}
+            'time': times,
+            'percentage': percentages}
 
 
-try:
-    from bokeh.palettes import viridis
-except ImportError:
-    palette = ['red', 'green', 'blue', 'yellow']
-else:
-    palette = viridis(10)
+def _watch(thread_id, log, interval='20ms', cycle='2s', omit=None,
+           stop=lambda: False):
+    interval = parse_timedelta(interval)
+    cycle = parse_timedelta(cycle)
 
-counter = itertools.count()
+    recent = create()
+    last = time()
+
+    while not stop():
+        if time() > last + cycle:
+            log.append((time(), recent))
+            recent = create()
+            last = time()
+        try:
+            frame = sys._current_frames()[thread_id]
+        except KeyError:
+            return
+
+        process(frame, None, recent, omit=omit)
+        sleep(interval)
 
 
-@toolz.memoize
-def color_of(x):
-    return palette[next(counter) % len(palette)]
+def watch(thread_id=None, interval='20ms', cycle='2s', maxlen=1000, omit=None,
+          stop=lambda: False):
+    if thread_id is None:
+        thread_id = get_thread_identity()
+
+    log = deque(maxlen=maxlen)
+
+    thread = threading.Thread(target=_watch,
+                              name='Profile',
+                              kwargs={'thread_id': thread_id,
+                                      'interval': interval,
+                                      'cycle': cycle,
+                                      'log': log,
+                                      'omit': omit,
+                                      'stop': stop})
+    thread.daemon = True
+    thread.start()
+
+    return log
+
+
+def get_profile(history, recent=None, start=None, stop=None, key=None):
+    now = time()
+    if start is None:
+        istart = 0
+    else:
+        istart = bisect.bisect_left(history, (start,))
+
+    if stop is None:
+        istop = None
+    else:
+        istop = bisect.bisect_right(history, (stop,)) + 1
+        if istop >= len(history):
+            istop = None  # include end
+
+    if istart == 0 and istop is None:
+        history = list(history)
+    else:
+        iistop = len(history) if istop is None else istop
+        history = [history[i] for i in range(istart, iistop)]
+
+    prof = merge(*toolz.pluck(1, history))
+
+    if not history:
+        return create()
+
+    if recent:
+        prof = merge(prof, recent)
+
+    return prof
+
+
+def plot_figure(data, **kwargs):
+    from bokeh.plotting import ColumnDataSource, figure
+    from bokeh.models import HoverTool
+
+    if 'states' in data:
+        data = toolz.dissoc(data, 'states')
+
+    source = ColumnDataSource(data=data)
+
+    fig = figure(tools='tap', **kwargs)
+    r = fig.quad('left', 'right', 'top', 'bottom', color='color',
+             line_color='black', line_width=2, source=source)
+
+    r.selection_glyph = None
+    r.nonselection_glyph = None
+
+    hover = HoverTool(
+        point_policy="follow_mouse",
+        tooltips="""
+            <div>
+                <span style="font-size: 14px; font-weight: bold;">Name:</span>&nbsp;
+                <span style="font-size: 10px; font-family: Monaco, monospace;">@name</span>
+            </div>
+            <div>
+                <span style="font-size: 14px; font-weight: bold;">Filename:</span>&nbsp;
+                <span style="font-size: 10px; font-family: Monaco, monospace;">@filename</span>
+            </div>
+            <div>
+                <span style="font-size: 14px; font-weight: bold;">Line number:</span>&nbsp;
+                <span style="font-size: 10px; font-family: Monaco, monospace;">@line_number</span>
+            </div>
+            <div>
+                <span style="font-size: 14px; font-weight: bold;">Line:</span>&nbsp;
+                <span style="font-size: 10px; font-family: Monaco, monospace;">@line</span>
+            </div>
+            <div>
+                <span style="font-size: 14px; font-weight: bold;">Time:</span>&nbsp;
+                <span style="font-size: 10px; font-family: Monaco, monospace;">@time</span>
+            </div>
+            <div>
+                <span style="font-size: 14px; font-weight: bold;">Percentage:</span>&nbsp;
+                <span style="font-size: 10px; font-family: Monaco, monospace;">@width</span>
+            </div>
+            """
+    )
+    fig.add_tools(hover)
+
+    fig.xaxis.visible = False
+    fig.yaxis.visible = False
+    fig.grid.visible = False
+
+    return fig, source
