@@ -8,10 +8,10 @@ import functools
 import gc
 from glob import glob
 import itertools
-import inspect
 import logging
 import logging.config
 import os
+import psutil
 import re
 import shutil
 import signal
@@ -34,20 +34,23 @@ import pytest
 import six
 
 import dask
-from toolz import merge, memoize
+from toolz import merge, memoize, assoc
 from tornado import gen, queues
 from tornado.gen import TimeoutError
 from tornado.ioloop import IOLoop
 
-from .client import default_client
-from .compatibility import PY3, iscoroutinefunction, Empty
+from .client import default_client, _global_clients, Client
+from .compatibility import PY3, Empty, WINDOWS, PY2
+from .comm.utils import offload
 from .config import initialize_logging
 from .core import connect, rpc, CommClosedError
 from .metrics import time
+from .process import _cleanup_dangling
 from .proctitle import enable_proctitle_on_children
 from .security import Security
 from .utils import (ignoring, log_errors, mp_context, get_ip, get_ipv6,
-                    DequeHandler, reset_logger_locks, sync)
+                    DequeHandler, reset_logger_locks, sync,
+                    iscoroutinefunction, thread_state)
 from .worker import Worker, TOTAL_MEMORY, _global_workers
 
 try:
@@ -64,6 +67,9 @@ logging_levels = {name: logger.level for name, logger in
                   if isinstance(logger, logging.Logger)}
 
 
+offload(lambda: None).result()  # create thread during import
+
+
 @pytest.fixture(scope='session')
 def valid_python_script(tmpdir_factory):
     local_file = tmpdir_factory.mktemp('data').join('file.py')
@@ -97,6 +103,7 @@ def cleanup_global_workers():
 @pytest.fixture
 def loop():
     del _global_workers[:]
+    _global_clients.clear()
     with pristine_loop() as loop:
         # Monkey-patch IOLoop.start to wait for loop stop
         orig_start = loop.start
@@ -126,6 +133,20 @@ def loop():
             is_stopped.wait()
     del _global_workers[:]
 
+    start = time()
+    while set(_global_clients):
+        sleep(0.1)
+        assert time() < start + 10
+
+    _cleanup_dangling()
+
+    if PY2:  # no forkserver, so no extra procs
+        for child in psutil.Process().children(recursive=True):
+            with ignoring(psutil.NoSuchProcess):
+                child.terminate()
+
+    _global_clients.clear()
+
 
 @pytest.fixture
 def loop_in_thread():
@@ -513,6 +534,75 @@ def check_active_rpc(loop, active_rpc_timeout=1):
     loop.run_sync(wait)
 
 
+@pytest.fixture
+def cluster_fixture(loop):
+    with cluster() as (scheduler, workers):
+        yield (scheduler, workers)
+
+
+@pytest.fixture
+def s(cluster_fixture):
+    scheduler, workers = cluster_fixture
+    return scheduler
+
+
+@pytest.fixture
+def a(cluster_fixture):
+    scheduler, workers = cluster_fixture
+    return workers[0]
+
+
+@pytest.fixture
+def b(cluster_fixture):
+    scheduler, workers = cluster_fixture
+    return workers[1]
+
+
+@pytest.fixture
+def client(loop, cluster_fixture):
+    scheduler, workers = cluster_fixture
+    with Client(scheduler['address'], loop=loop) as client:
+        yield client
+
+
+@pytest.fixture
+def client_secondary(loop, cluster_fixture):
+    scheduler, workers = cluster_fixture
+    with Client(scheduler['address'], loop=loop) as client:
+        yield client
+
+
+@contextmanager
+def tls_cluster_context(worker_kwargs=None, scheduler_kwargs=None,
+                        security=None, **kwargs):
+    security = security or tls_only_security()
+    worker_kwargs = assoc(worker_kwargs or {}, 'security', security)
+    scheduler_kwargs = assoc(scheduler_kwargs or {}, 'security', security)
+
+    with cluster(worker_kwargs=worker_kwargs,
+                 scheduler_kwargs=scheduler_kwargs,
+                 **kwargs) as (s, workers):
+        yield s, workers
+
+
+@pytest.fixture
+def tls_cluster(loop, security):
+    with tls_cluster_context(security=security) as (scheduler, workers):
+        yield (scheduler, workers)
+
+
+@pytest.fixture
+def tls_client(tls_cluster, loop, security):
+    s, workers = tls_cluster
+    with Client(s['address'], security=security, loop=loop) as client:
+        yield client
+
+
+@pytest.fixture
+def security():
+    return tls_only_security()
+
+
 @contextmanager
 def cluster(nworkers=2, nanny=False, worker_kwargs={}, active_rpc_timeout=1,
             scheduler_kwargs={}):
@@ -568,7 +658,13 @@ def cluster(nworkers=2, nanny=False, worker_kwargs={}, active_rpc_timeout=1,
 
             start = time()
             try:
-                with rpc(saddr) as s:
+                try:
+                    security = scheduler_kwargs['security']
+                    rpc_kwargs = {'connection_args': security.get_connection_args('client')}
+                except KeyError:
+                    rpc_kwargs = {}
+
+                with rpc(saddr, **rpc_kwargs) as s:
                     while True:
                         ncores = loop.run_sync(s.ncores)
                         if len(ncores) == nworkers:
@@ -584,8 +680,9 @@ def cluster(nworkers=2, nanny=False, worker_kwargs={}, active_rpc_timeout=1,
                 logger.debug("Closing out test cluster")
 
                 loop.run_sync(lambda: disconnect_all([w['address'] for w in workers],
-                                                     timeout=0.5))
-                loop.run_sync(lambda: disconnect(saddr, timeout=0.5))
+                                                     timeout=0.5,
+                                                     rpc_kwargs=rpc_kwargs))
+                loop.run_sync(lambda: disconnect(saddr, timeout=0.5, rpc_kwargs=rpc_kwargs))
 
                 scheduler.terminate()
                 scheduler_q.close()
@@ -618,15 +715,20 @@ def cluster(nworkers=2, nanny=False, worker_kwargs={}, active_rpc_timeout=1,
             else:
                 client.close()
 
-    assert not ws
+    start = time()
+    while list(ws):
+        sleep(0.01)
+        assert time() < start + 1, 'Workers still around after one second'
 
 
 @gen.coroutine
-def disconnect(addr, timeout=3):
+def disconnect(addr, timeout=3, rpc_kwargs=None):
+    rpc_kwargs = rpc_kwargs or {}
+
     @gen.coroutine
     def do_disconnect():
         with ignoring(EnvironmentError, CommClosedError):
-            with rpc(addr) as w:
+            with rpc(addr, **rpc_kwargs) as w:
                 yield w.terminate(close=True)
 
     with ignoring(TimeoutError):
@@ -634,8 +736,8 @@ def disconnect(addr, timeout=3):
 
 
 @gen.coroutine
-def disconnect_all(addresses, timeout=3):
-    yield [disconnect(addr, timeout) for addr in addresses]
+def disconnect_all(addresses, timeout=3, rpc_kwargs=None):
+    yield [disconnect(addr, timeout, rpc_kwargs) for addr in addresses]
 
 
 def slow(func):
@@ -717,17 +819,11 @@ def end_cluster(s, workers):
     s.stop()
 
 
-def iscoroutinefunction(f):
-    if sys.version_info >= (3, 5) and inspect.iscoroutinefunction(f):
-        return True
-    return False
-
-
 def gen_cluster(ncores=[('127.0.0.1', 1), ('127.0.0.1', 2)],
                 scheduler='127.0.0.1', timeout=10, security=None,
                 Worker=Worker, client=False, scheduler_kwargs={},
                 worker_kwargs={}, client_kwargs={}, active_rpc_timeout=1,
-                config={}):
+                config={}, check_new_threads=True):
     from distributed import Client
     """ Coroutine test with small cluster
 
@@ -739,11 +835,6 @@ def gen_cluster(ncores=[('127.0.0.1', 1), ('127.0.0.1', 2)],
         start
         end
     """
-    del _global_workers[:]
-
-    reset_config()
-
-    dask.config.set({'distributed.comm.timeouts.connect': '5s'})
     worker_kwargs = merge({'memory_limit': TOTAL_MEMORY, 'death_timeout': 5},
                           worker_kwargs)
 
@@ -752,6 +843,13 @@ def gen_cluster(ncores=[('127.0.0.1', 1), ('127.0.0.1', 2)],
             func = gen.coroutine(func)
 
         def test_func():
+            del _global_workers[:]
+            _global_clients.clear()
+            active_threads_start = set(threading._active)
+
+            reset_config()
+
+            dask.config.set({'distributed.comm.timeouts.connect': '5s'})
             # Restore default logging levels
             # XXX use pytest hooks/fixtures instead?
             for name, level in logging_levels.items():
@@ -765,6 +863,7 @@ def gen_cluster(ncores=[('127.0.0.1', 1), ('127.0.0.1', 2)],
                     @gen.coroutine
                     def coro():
                         with dask.config.set(config):
+                            s = False
                             for i in range(5):
                                 try:
                                     s, ws = yield start_cluster(
@@ -774,9 +873,11 @@ def gen_cluster(ncores=[('127.0.0.1', 1), ('127.0.0.1', 2)],
                                 except Exception as e:
                                     logger.error("Failed to start gen_cluster, retryng", exc_info=True)
                                 else:
+                                    workers[:] = ws
+                                    args = [s] + workers
                                     break
-                            workers[:] = ws
-                            args = [s] + workers
+                            if s is False:
+                                raise Exception("Could not start cluster")
                             if client:
                                 c = yield Client(s.address, loop=loop, security=security,
                                                  asynchronous=True, **client_kwargs)
@@ -807,22 +908,43 @@ def gen_cluster(ncores=[('127.0.0.1', 1), ('127.0.0.1', 2)],
 
                     result = loop.run_sync(coro, timeout=timeout * 2 if timeout else timeout)
 
-            for w in workers:
-                if getattr(w, 'data', None):
-                    try:
-                        w.data.clear()
-                    except EnvironmentError:
-                        # zict backends can fail if their storage directory
-                        # was already removed
-                        pass
-                    del w.data
-            DequeHandler.clear_all_instances()
-            for w in _global_workers:
-                w = w()
-                w._close(report=False, executor_wait=False)
-                if w.status == 'running':
-                    w.close()
-            del _global_workers[:]
+                for w in workers:
+                    if getattr(w, 'data', None):
+                        try:
+                            w.data.clear()
+                        except EnvironmentError:
+                            # zict backends can fail if their storage directory
+                            # was already removed
+                            pass
+                        del w.data
+                DequeHandler.clear_all_instances()
+                for w in _global_workers:
+                    w = w()
+                    w._close(report=False, executor_wait=False)
+                    if w.status == 'running':
+                        w.close()
+                del _global_workers[:]
+
+            if PY3 and not WINDOWS and check_new_threads:
+                start = time()
+                while True:
+                    bad = [t for t, v in threading._active.items()
+                           if t not in active_threads_start and
+                          "Threaded" not in v.name and
+                          "watch message" not in v.name]
+                    if not bad:
+                        break
+                    else:
+                        sleep(0.01)
+                    if time() > start + 5:
+                        from distributed import profile
+                        tid = bad[0]
+                        thread = threading._active[tid]
+                        call_stacks = profile.call_stack(sys._current_frames()[tid])
+                        assert False, (thread, call_stacks)
+            _cleanup_dangling()
+            with ignoring(AttributeError):
+                del thread_state.on_event_loop_thread
             return result
 
         return test_func
@@ -967,7 +1089,7 @@ def assert_can_connect(addr, timeout=None, connection_args=None):
     within the given *timeout*.
     """
     if timeout is None:
-        timeout = 0.2
+        timeout = 0.5
     comm = yield connect(addr, timeout=timeout,
                          connection_args=connection_args)
     comm.abort()
@@ -980,7 +1102,7 @@ def assert_cannot_connect(addr, timeout=None, connection_args=None, exception_cl
     within the given *timeout*.
     """
     if timeout is None:
-        timeout = 0.2
+        timeout = 0.5
     with pytest.raises(exception_class):
         comm = yield connect(addr, timeout=timeout,
                              connection_args=connection_args)
